package de.mdelab.mlsdm.interpreter.patternMatcher.patternPartBased;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.Activity;
import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.MapEntryLink;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.expressions.ExpressionInterpreterManager;
import de.mdelab.sdm.interpreter.core.facade.MetamodelFacadeFactory;
import de.mdelab.sdm.interpreter.core.notifications.NotificationEmitter;
import de.mdelab.sdm.interpreter.core.patternmatcher.MatchingStrategy;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.ECheckResult;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.EMatchType;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.MatchStoryPatternObjectTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.PatternPart;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.PatternPartBasedMatcher;
import de.mdelab.sdm.interpreter.core.variables.Variable;
import de.mdelab.sdm.interpreter.core.variables.VariablesScope;

public class MLSDMNACPatternPart extends PatternPart<AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>
{

	private final StoryPattern					nacPattern;

	private final AbstractStoryPatternObject[]	storyPatternObjects;

	public MLSDMNACPatternPart(
			final PatternPartBasedMatcher<?, ?, ?, ?, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, ?, MLExpression> patternMatcher,
			final StoryPattern nacPattern)
	{
		super(patternMatcher);

		assert nacPattern != null;
		assert nacPattern.getContainingPattern() != null;

		this.nacPattern = nacPattern;

		/*
		 * Collect story pattern objects connected to this NAC pattern
		 */
		final Set<AbstractStoryPatternObject> spos = new HashSet<AbstractStoryPatternObject>();

		for (final AbstractStoryPatternLink link : nacPattern.getStoryPatternLinks())
		{
			if (link.getSource().getStoryPattern() == nacPattern.getContainingPattern())
			{
				spos.add(link.getSource());

				assert link.getTarget().getStoryPattern() == nacPattern || link instanceof MapEntryLink
						&& ((MapEntryLink) link).getValueTarget().getStoryPattern() == nacPattern;
			}
			else if (link.getTarget().getStoryPattern() == nacPattern.getContainingPattern())
			{
				spos.add(link.getTarget());

				assert link.getSource().getStoryPattern() == nacPattern || link instanceof MapEntryLink
						&& ((MapEntryLink) link).getValueTarget().getStoryPattern() == nacPattern;
			}

			if (link instanceof MapEntryLink
					&& ((MapEntryLink) link).getValueTarget().getStoryPattern() == nacPattern.getContainingPattern())
			{
				spos.add(((MapEntryLink) link).getValueTarget());
			}
		}

		this.storyPatternObjects = spos.toArray(new AbstractStoryPatternObject[spos.size()]);
	}

	@Override
	public AbstractStoryPatternObject[] getStoryPatternObjects()
	{
		return this.storyPatternObjects;
	}

	@Override
	public EMatchType getMatchType()
	{
		return EMatchType.MANDATORY;
	}

	@Override
	public ECheckResult check() throws SDMException
	{
		/*
		 * If all story pattern objects are bound
		 */
		for (final AbstractStoryPatternObject spo : this.nacPattern.getContainingPattern().getStoryPatternObjects())
		{
			if (!this.patternMatcher.isBound(spo))
			{
				return ECheckResult.UNKNOWN;
			}
		}

		/*
		 * Try to find a match for all contained pattern parts
		 */

		/*
		 * Create a new pattern matcher for the sub pattern
		 */
		try
		{
			@SuppressWarnings("unchecked")
			final MLSDMPatternPartBasedMatcher subPatternMatcher = new MLSDMPatternPartBasedMatcher(
					this.nacPattern,
					(VariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) this.patternMatcher
							.getVariablesScope(),
					(MatchingStrategy<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) this.patternMatcher
							.getMatchingStrategy(),
					(MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) this.patternMatcher
							.getFacadeFactory(),
					(ExpressionInterpreterManager<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) this.patternMatcher
							.getExpressionInterpreterManager(),
					(NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) this.patternMatcher
							.getNotificationEmitter())
			{
				@Override
				protected boolean analyzeStoryPatternObjects() throws SDMException
				{
					/*
					 * analyze all story pattern objects
					 */
					for (final AbstractStoryPatternObject spo : new HashSet<AbstractStoryPatternObject>(this.unboundSPO))
					{
						/*
						 * All story pattern objects outside of the NAC pattern
						 * are treated as bound objects.
						 */
						final boolean bound = !MLSDMNACPatternPart.this.nacPattern.getStoryPatternObjects().contains(spo);

						if (bound)
						{
							final Variable<EClassifier> variable = this.getVariablesScope().getVariable(spo.getName());

							if (variable == null)
							{
								throw new UnsupportedOperationException();
							}

							/*
							 * Check that the type of the existing variable
							 * matches the type of the story pattern object,
							 * check isomorphism condition and object
							 * constraints
							 */
							if (variable.getValue() != null
									&& this.checkTypeConformance(this.spoFacade.getClassifier(spo), variable.getValue())
									&& this.checkIsomorphism(variable.getValue())
									&& this.checkStoryPatternObjectConstraints(spo, variable.getValue()))
							{
								/*
								 * push matchSpoTransaction onto stack
								 */
								final MatchStoryPatternObjectTransaction<AbstractStoryPatternObject> matchStoryPatternObjectTransaction = new MatchStoryPatternObjectTransaction<AbstractStoryPatternObject>(
										spo, variable.getValue(), this.unboundSPO, this.boundSPO, this.boundInstanceObjects);
								matchStoryPatternObjectTransaction.commit();
								this.matchingStack.push(matchStoryPatternObjectTransaction);

								this.getNotificationEmitter().storyPatternObjectBound(spo, variable.getValue(), this.getVariablesScope(),
										this);
							}
							else
							{
								throw new UnsupportedOperationException();
							}
						}
						else
						{
							/*
							 * Delete existing variables of unbound story
							 * pattern objects
							 */
							this.getVariablesScope().deleteVariable(this.spoFacade.getName(spo));
						}
					}

					return true;
				}
			};

			MATCHLOOP: while (subPatternMatcher.findNextMatch())
			{
				/*
				 * Check isomorphism
				 */
				StoryPattern pattern = this.nacPattern;

				/*
				 * Collect bound objects of this pattern and its parent
				 * patterns.
				 */
				final Set<Object> boundObjects = new HashSet<Object>();

				do
				{
					for (final AbstractStoryPatternObject spo : pattern.getStoryPatternObjects())
					{
						final Variable<EClassifier> var = subPatternMatcher.getVariablesScope().getVariable(spo.getName());

						assert var != null;
						assert var.getValue() != null;

						if (boundObjects.contains(var.getValue()))
						{
							continue MATCHLOOP;
						}
						else
						{
							boundObjects.add(var.getValue());
						}
					}

					pattern = pattern.getContainingPattern();
				}
				while (pattern != null);

				return ECheckResult.FAIL;
			}

			/*
			 * No match for the NAC found
			 */
			return ECheckResult.OK;
		}
		catch (final Exception ex)
		{
			throw new RuntimeException(ex);
		}

	}

	@Override
	public boolean match(final MatchState matchState) throws SDMException
	{
		throw new UnsupportedOperationException();
	}

	@Override
	public void createObjects() throws SDMException
	{
		// Do nothing
	}

	@Override
	public void createLinks() throws SDMException
	{
		// Do nothing
	}

	@Override
	public void destroyObjects()
	{
		// Do nothing
	}

	@Override
	public int calculateMatchingCost()
	{
		return PatternPart.MATCHING_NOT_POSSIBLE;
	}

	@Override
	public MatchState createMatchState()
	{
		return null;
	}

	@Override
	public void destroyLinks(final Map<AbstractStoryPatternObject, Object> deletedObjects)
	{
		// Do nothing
	}

}
